#!/usr/bin/env python3
"""
Printing routines for seaplan
"""
# STANDARD LIBRARIES
from datetime import datetime, timedelta
import math

from .version import __version__

# CONSTANTS
comment_len = 40


def event_table(event_list, parameter_file, params, format="RST"):
    """
    Print a table in MultiMarkDown or reStructuredText
    
    :param event_list: list of Events
    :param parameter_file:
    :param params:
    :param format: 'RST' or 'MMD'

    Should add HTML output
    """
    assert format in ['RST', 'MMD']
        
    show_comments = params['printout']['show_comments']
    t = _table_header(parameter_file, params, format)
    prev_arrival = event_list[0]['arrival']['time']
    day_mesg = make_day_mesg(prev_arrival, show_comments, format, True)
    trans_accum = dict(h=0., nm=0.)
    for event in event_list:
        if event['hours'] or params['printout']['print_waypoints']:
            arrival = event['arrival']['time']
            departure = event['departure']['time']
            if arrival.day != prev_arrival.day:
                day_mesg = make_day_mesg(arrival, show_comments, format)
            if (params['printout']['print_past']
                or (departure > datetime.now())):
                    if day_mesg:
                        t += day_mesg
                        day_mesg = ''
                    t += _table_row(event, trans_accum, show_comments, format)
            prev_arrival = arrival
            trans_accum = dict(h=0., nm=0.)
        else:
            trans_accum['h'] += event['transit_hours']
            trans_accum['nm'] += event['dist_from_previous']
    t += _table_footer(show_comments, format)
    return t


# def _parse_rst(text: str) -> docutils.nodes.document:
#     parser = docutils.parsers.rst.Parser()
#     components = (docutils.parsers.rst.Parser,)
#     settings = docutils.frontend.OptionParser(components=components).\
#         get_default_values()
#     document = docutils.utils.new_document('<rst-doc>', settings=settings)
#     parser.parse(text, document)
#     return document


def make_day_mesg(the_time, show_comments, form="multimarkdown", no_top=False):
    if form == "multimarkdown":
        t = "|||||| {:63s}   ||".format(the_time.strftime("*%A, %d %B*"))
    else:
        # reStructuredText
        if show_comments:
            if no_top:
                t = ""
            else:
                t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+-" + comment_len*"-" + "-|\n"
            t += "|    {:85s}    ".format(the_time.strftime("*%A, %d %B*"))\
                + comment_len*" " + " |\n"
        else:
            if no_top:
                t = ""
            else:
                t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+\n"
            t += "|    {:85s}  |\n".format(the_time.strftime("*%A, %d %B*"))
    return t


def _table_header(parameter_file, params, form='multimarkdown'):
    t = _header1(params['description']['title'], form)
    t += f"seaplan version {__version__}\n\n"
    t += f"Ran at {datetime.now().strftime('%Y-%m-%d %H:%M local')}\n\n"
    t += _header2("Notes", form)
    for note in params['description']['notes']:
        t += f"- {note}\n"
    t += "\n"
    t += _header2("Parameters", form)
    t += f'- **Station file**: "{params["station_file"]["name"]}"\n'
    t += f'- **Parameter file**: "{parameter_file}"\n'
    t += f'\n'
    t += '  - **Boat speed (kts)**: {:d}\n'.format(params['timing']['ship_speed.kn_i'])
    t += '  - **Boat latency (h)**: {:g}\n'.format(params['timing']['ship_latency.h'])
    t += '  - **Default Action Times (hours)**:\n'
    t += '\n'
    for action, values in params['timing']['default_action_times.h'].items():
        t += f'    - **{action}**: '
        for name, hours in values.items():        
            t += f'"{name}": {hours}, '
        t += '\n'
    t += '\n'
    if params['timing']['ignore_depart_arrive_times']:
        t += '\nIGNORING ALL "arrival_time"S AND "departure_time"S IN PARAMETER FILE!!!\n'
    if form == 'multimarkdown':
        if not params['printout']['show_comments']:
            t += "\n"
            t += "|    EVENT       ||  POSITION           ||   TRANSIT   |STATION TIME (Z)||\n"
            t += " Station | Action |   Lat    |   Lon    | hours(nmiles)| Arrive | Depart \n"
            t += "---------|--------|---------:|---------:|-------------:|:------:|:------:\n"
        else:
            t += "\n"
            t += "|    EVENT       ||        POSITION     ||    TRANSIT  |STATION TIME (Z)|| COMMENTS\n"
            t += " Station | Action |   Lat    |   Lon    | hours(nmiles)| Arrive | Depart | \n"
            t += "---------|--------|---------:|---------:|-------------:|:------:|:------:|---------------------\n"
    else:
        # reStructuredText
        if not params['printout']['show_comments']:
            t += "\n"
            t += "+--------------------+-----------------------+--------------+-------------------------------+\n"
            t += "| EVENT              | POSITION              | TRANSIT      | STATION TIME                  |\n"
            t += "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+\n"
            t += "| Station | Action   | Lat       | Lon       | hours(nmiles)| Hours | Arrive    | Depart    |\n"
            t += "+=========+==========+===========+===========+==============+=======+===========+===========+\n"
        else:
            t += "\n"
            t += "+--------------------+-----------------------+--------------+-------------------------------+-"+comment_len*"-"+"-+\n"
            t += "| EVENT              | POSITION              | TRANSIT      | STATION TIME          | COMMENTS "+(comment_len-9)*" "+" |\n"
            t += "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+-"+comment_len*"-"+"-+\n"
            t += "| Station | Action   | Lat       | Lon       | hours(nmiles)| Hours | Arrive    | Depart    | "+comment_len*" "+" |\n"
            t += "+=========+==========+===========+===========+==============+=======+===========+===========+="+comment_len*"="+"=+\n"
    return t

def _header1(text, form):
    if form=="multimarkdown":
        t = f"# {text}\n\n"
    else:
        t = f"{text}\n"
        t += "-" * len(text)
        t += "\n\n"
    return t

def _header2(text, form):
    if form=="multimarkdown":
        t = f"## {text}\n\n"
    else:
        t = f"{text}\n"
        t += "=" * len(text)
        t += "\n\n"
    return t

def _table_row(event, trans_accum, show_comments, form="multimarkdown",
               debug=False):
    sta = event['station']
    act = event.get('action', '')
    lat = __deg_as_degmin(event['lat'])
    lon = __deg_as_degmin(event['lon'])
    trn = event['transit_hours'] + trans_accum['h']
    dst = event['dist_from_previous'] + trans_accum['nm']
    arr = time_str(event['arrival'])
    dep = time_str(event['departure'])
    evtt = (event['departure']['time']-event['arrival']['time'])/timedelta(hours=1)
    evtt_s = '{:02d}:{:02d}'.format(int(math.floor(evtt)), int(60*(evtt-math.floor(evtt))))
    cmt = event['comment']
    if debug:
        print(f"comment={event['comment']}")
    if form == "multimarkdown":
        if not show_comments:
            return "{:9.9s}| {:6.6s} |{:>9s} |{:>9s} | {:4.1f} ({:5.1f}) |{:^7s}|{:^7s}\n".format(
                 sta, act, lat, lon, trn, dst, arr, dep)
        else:
            return "{:9.9s}| {:6.6s} |{:>9s} |{:>9s} | {:4.1f} ({:5.1f}) |{:^7s}|{:^7s} | {}\n".format(
                 sta, act, lat, lon, trn, dst, arr, dep, cmt)
    else:
        # reStructuredText
        if not show_comments:
            t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+\n"
            t += "| {:7.7s} | {:8.8s} | {:>9s} | {:>9s} | {:4.1f} ({:5.1f}) | {:^5s} | {:^9s} | {:^9s} |\n".format(
                 sta, act, lat, lon, trn, dst, evtt_s, arr, dep)
            return t
        else:
            t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+-"+comment_len*"-"+"-+\n"
            t += "| {:7.7s} | {:8.8s} | {:>9s} | {:>9s} | {:4.1f} ({:5.1f}) | {:^5s} | {:^9s} | {:^9s} | {:" + f"{comment_len:d}.{comment_len:d}s"+"} |\n"
            return t.format(sta, act, lat, lon, trn, dst, evtt_s, arr, dep, cmt)


def _table_footer(show_comments, form="multimarkdown"):
    if form == "multimarkdown":
        t = "\n"
        t += "To make the above table pretty, convert to HTML using MMD and\n"
        t += "replace the <table> tag in the resulting HTML file with:\n"
        t += '<table border="1" width="100%" cellspacing="0" cellpadding=6>\n'
        t += "\n"
    else:
        if show_comments:
            t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+-" + 40*"-" + "-+\n"
        else:
            t = "+---------+----------+-----------+-----------+--------------+-------+-----------+-----------+\n"
    return t


def time_str(the_time):
    """
    return time string
    
    :param the_time: dict with 'time' and 'forced' keys

    The output text has different emphasis based on whether it is
    the input time (normal), a forced time in the future (**bold**),
    or a forced time in the past (*italic*)
    """
    emphasis = ''
    if the_time['forced']:
        if the_time['time'] < datetime.now():  # Already passed, no big deal
            emphasis = '*'
        else:                            # Not yet passed, **CONSTRAINT**
            emphasis = '**'
    time_s = emphasis+the_time['time'].strftime("%H:%M") + emphasis
    return time_s


def print_used_stations(stations, used_stations):
    """
    Print information about which stations were used

    I think this is inefficiently written (no need for sta_dict?)
    :param stations: list of station dictionaries
    :param used_stations: list of the stations visited, in order
    """
    sta_dict = dict()
    for us in used_stations:
        sta_dict[us] = sta_dict.get(us, 0) + 1
    print('{:d} of {:d} stations in the station file were used:'.format(
        len(sta_dict), len(stations)))

    # List used stations
    print("Used:")
    __print_by_cols(sorted(sta_dict))
    # List unused stations
    print("Unused:")
    unused = []
    for name in stations.keys():
        if name not in sta_dict.keys():
            unused.append(name)
    if len(unused) == 0:
        print("  None")
    else:
        __print_by_cols(sorted(unused))

    # If multiple occurences of a station or stations, warn and inform
    iRepeated = 0
    for sta_name in sorted(sta_dict.keys()):
        if sta_dict[sta_name] > 1:
            iRepeated += 1
    if iRepeated:
        if iRepeated == 1:
            print('1 repeated used station:')
        else:
            print('{:d} of the stations were repeated:'.format(iRepeated))
        for sta_name in sorted(sta_dict.keys()):
            if sta_dict[sta_name] > 1:
                print("  {} : {:d} times".format(sta_name, sta_dict[sta_name]))
    else:
        print("No repeated used stations")
    print('')


def __print_by_cols(list, ncols=10):
    " print given list as rows of a given number of columns (10 by default) "
    i = 1
    for obj in list:
        if i == 1:
            print('  ', end='')
        print('{:>6s},'.format(obj), end='')
        i += 1
        if i > ncols:
            print("")
            i = 1
    if i != 1:
        print("")


def __deg_as_degmin(degrees, precision=1):
    """
    Return text string containing degrees and minutes

    precision is number of digits after the decimal place to return
    for the minutes field
    """
    if not isinstance(precision, int):
        print('PRECISION IS NOT AN INTEGER')
        return
    if precision < 0:
        print('PRECISION < 0!')
        return
    deg_int = math.trunc(degrees)
    deg_min = abs(60 * (degrees - deg_int))
    field_width = precision + 3
    if precision == 0:
        field_width = 2
    else:
        field_width = precision + 3
    fmt_str = "{:d}d{:0" + "{:d}.{:d}".format(field_width, precision) + "f}m"
    return fmt_str.format(deg_int, deg_min)



"""
Validate information files

"""
# Standard library modules
import json
# import pprint
import os.path
import sys
import pkg_resources

# Non-standard modules
import jsonschema
import jsonref
import yaml

lazy_errors = True   # Can report multiple errors, but less stable
VALID_FORMATS = ['JSON', 'YAML']


def validate(filename, format=None, verbose=True, quiet=False):
    """
    Validates a YAML or JSON file against schema
    format: "JSON" or "YAML"

    if format is not provided, tries to figure it out from the
    filename suffix
    """

    if quiet:
        verbose = False

    SCHEMA_FILE = pkg_resources.resource_filename(
        'seaplan', f'data/seaplan.schema.json')

    if verbose:
        print(f'  Validating {filename} against '
              f'{os.path.basename(SCHEMA_FILE)}')

    instance = read_json_yaml(filename, format=format)

    base_path = os.path.dirname(SCHEMA_FILE)
    base_uri = f'file://{base_path}/'

    # LOAD SCHEMA FILE
    with open(SCHEMA_FILE, 'r') as f:
        try:
            schema = jsonref.loads(f.read(), base_uri=base_uri,
                                   jsonschema=True)
        except json.decoder.JSONDecodeError as e:
            print(f'JSONDecodeError: Error loading JSON schema file: '
                  f'{SCHEMA_FILE}')
            print(str(e))
            return
        except:
            print(f"Error loading JSON schema file: {SCHEMA_FILE}")
            print(sys.exc_info()[1])
            return
    # VALIDATE SCHEMA FILE
    try:
        if verbose:
            print('\tTesting schema ...', end='')

        v = jsonschema.Draft4Validator(schema)

        if verbose:
            print('OK, instance ...', end='')
    except jsonschema.ValidationError as e:
        if quiet:
            # IF HAVE TO PRINT ERROR MESSAGE, PRINT INTRO TOO
            print(f'instance = {filename}')
        else:
            print('')
        print('\t' + e.message)
    except:
        print("schema failed and I don't know why")
    # VALIDATE PARAMETER FILE
    try:
        if not v.is_valid(instance):
            if quiet:
                # IF HAVE TO PRINT ERROR MESSAGE, PRINT INTRO TOO
                print(f'instance = {filename}')
            else:
                print('')
            if lazy_errors:
                # Lazily report all errors in the instance
                # ASSUMES SCHEMA IS DRAFT-04 (I couldn't get it to work
                # otherwise)
                try:
                    for error in v.iter_errors(instance):
                        print('\t\t', end='')
                        try:
                            for elem in error.path:
                                print(f"['{elem}']", end='')
                        except:
                            print("Error in error.path")
                        print(f': {error.message}')
                except:
                    print('Error in v.iter_errors(instance)')
            else:
                # Just report first error and bail
                error = next(v.iter_errors(instance))
                print('\t\t', end='')
                for elem in error.path:
                    print(f"['{elem}']", end='')
                print(f': {error.message}')
            print('\tFAILED')
        else:
            if not quiet:
                print('OK')
    except jsonschema.ValidationError as e:
        if quiet:
            # IF HAVE TO PRINT ERROR MESSAGE, PRINT INTRO TOO
            print(f'instance = {filename}')
        else:
            print('')
        print('\t' + e.message)
    except jsonschema.exceptions.UnknownType as e:
        print(e)
    return


def read_json_yaml(filename, format=None, debug=False):
    """ Reads a JSON or YAML file """
    if not format:
        format = get_information_file_format(filename)

    with open(filename, 'r') as f:
        if format == 'YAML':
            try:
                element = yaml.safe_load(f)
            except:
                print(f"Error loading YAML file: {filename}")
                print(sys.exc_info()[1])
                return
        else:
            try:
                element = json.load(filename)
            except JSONDecodeError as e:
                print(f"JSONDecodeError: Error loading JSON file: {filename}")
                print(str(e))
                return
            except:
                print(f"Error loading JSON file: {filename}")
                print(sys.exc_info()[1])
                return

    return element


def get_information_file_format(filename):
    """
    Determines if the information file is in JSON or YAML format

    Assumes that the filename is "*.{FORMAT}
    """

    format = filename.split('.')[-1].upper()
    if format in VALID_FORMATS:
        return format
    print('Unknown format: {format}')
    sys.exit(1)


def _console_script(argv=None):
    """
    Validate a seaplan information file

    Validates a file named *.seaplan.json or *.seaplan.yaml against
    schema.seaplan.json.
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(prog='seaplan-validate', description=__doc__)
    parser.add_argument('info_file', help='Information file')
    parser.add_argument('-f', '--format', choices=VALID_FORMATS,
                        default=None,
                        help='Forces information file to be interpreted as '
                             '"JSON" or "YAML"(overrides interpreting from '
                             'filename)')
    parser.add_argument('-v', '--verbose', action="store_true",
                        help='increase output verbosiy')
    args = parser.parse_args()

    validate(args.info_file, format=args.format, verbose=args.verbose)



#!/usr/bin/env python3
"""
Class to plot GMT-like bathymetry maps
"""
import warnings

import numpy as np
from scipy.io import netcdf
from matplotlib import pyplot as plt
import cartopy.crs as ccrs
from cartopy.feature import NaturalEarthFeature, GSHHSFeature, COASTLINE
from pylab import cm
import xarray as xr


class BathyMap():
    """
    Create a bathymetry map figure and axis
    """
    def __init__(self, map_extent, bathy_file,
                 grid_x=None, grid_y=None, intens_file=None):
        """
        Set up a bathymetric map

        :param map_extent [list]: [minlon,maxlon,minlat,maxlat]
        :param bathy_file [str]: name of netcdf bathymetry file
        :param intens_file: name of netcdf intensity file
        :kind intens_file: str, optional
        :param grid_x: x grid spacing for the plot axis
        :param grid_y: y grid spacing for the plot axis
        """
        self.map_extent = map_extent
        self.lon, self.lat, self.z = self._read_bathy_map(bathy_file)

        # set up figure and axes
        self.fig = plt.figure()
        self.ax = plt.axes(projection=ccrs.Mercator())
        self.ax.set_extent(self.map_extent)
        if grid_x:
            min_x = grid_x * np.floor(min(self.lon) / grid_x)
            grid_x = np.arange(min_x, max(self.lon), grid_x)
        if grid_y:
            min_y = grid_y * np.floor(min(self.lat) / grid_y)
            grid_y = np.arange(min_y, max(self.lat), grid_y)
        gl = self.ax.gridlines(xlocs=grid_x, ylocs=grid_y, draw_labels=True)
        gl.top_labels = False
        gl.right_labels = False

    def _read_bathy_map(self, fname):
        """
        Read netcdf grid file
        
        :param fname: filename
        """
        if not isinstance(fname, str):
            if fname is True:
                print('Sorry, ETOPO isn"t working')
            return self.map_extent[:2], self.map_extent[2:4], None
            
            # etopo = xr.open_dataset('etopo1.nc')
            # # Slice desired coordinates.
            # etopo = etopo.sel(x=slice(self.map_extent[0],self.map_extent[1]),
            #                   y=slice(self.map_extent[2],self.map_extent[3]))  
            # # Return sliced variables.
            # return etopo.x, etopo.y, etopo.z
        # Open ETOPO1 file and slice desired area.
        
        bathy = xr.open_dataset(fname)
        # Slice desired coordinates.
        bathy = bathy.sel(x=slice(self.map_extent[0],self.map_extent[1]),
                          y=slice(self.map_extent[2],self.map_extent[3]))  
        # Return sliced variables.
        return bathy.x, bathy.y, bathy.z
        
        # return lon[ixmin:ixmax], lat[iymin:iymax], z[iymin:iymax, ixmin:ixmax]

    def plot_image(self, cmap=cm.jet, pastel=False, shaded=False, intens_file=None, **kwargs):
        """
        Plot the bathymetric image

        :param cmap: colormap (e.g matplotlib.colors.LinearSegmentedColormap
              instance)
        :param pastel: lighten colors to make pastel
        :param shaded: [bool] apply shading
        :param intens_file: file containing intensities to plot (generally for shading)
        :param **kwargs: arguments to pass to hillshade (scale, azdeg, altdeg)
        """
        intens = None
        if intens_file is not None:
            lon, lat, intens = self._read_bathy_map(intens_file)
            if not (len(lon) == len(self.lon)) or not (len(lat) == len(self.lat)):
                warnings.warn('intensity lon/lat values different length than in bathy map')
            elif not all(abs(lon - self.lon) < ((lon[1] - lon[0]) / 10)) \
                    or not all(abs(lat - self.lat) < ((lat[1] - lat[0]) / 10)):
                warnings.warn('intensity lon/lat values do not match bathy map')
        if intens is not None:
            z_shade = _set_shade(self.z, intensity=intens, cmap=cmap, **kwargs)
        else:
            z_shade = _set_shade(self.z, cmap=cmap, shaded=shaded, **kwargs)
        if pastel:
            z_shade = 0.5 * z_shade + 0.5

        self.ax.imshow(z_shade, origin='lower', extent=self.map_extent,
                       transform=ccrs.PlateCarree())

    def plot_contours(self, levels=500, linewidth=1, color='k'):
        """
        Plot the bathymetric contours

        :param levels: list of contours, or contour interval (m)
        :param linewidth: contour linewidth (1)
        :param colors: contour line color ('k')
        """
        if not isinstance(levels, list):
            interval = levels
            min_level = interval * np.floor(np.nanmin(self.z)/interval)
            levels = np.arange(min_level, np.nanmax(self.z), interval)
        plt.contour(self.lon, self.lat, self.z,
                    levels, colors=color,
                    linestyles='solid', linewidths=linewidth,
                    transform=ccrs.PlateCarree())

    def show(self):
        """
        Show the plot on the screen
        """
        plt.show()

    def plot_coastlines(self, resolution=None):
        """
        Plot coastlines

        :param resolution: what resolution coastlines to include.  Must
            correspond to a NaturalEarth resolution ('10m', '50m', '110m')
            or a GSHSS resolution ('auto', 'low', 'high', 'full'...)
        """
        NE_coast_resolutions = ['10m', '50m', '110m']
        GSHSS_coast_resolutions = ['auto', 'coarse', 'low',
                                   'intermediate', 'high',
                                   'full']
        coast = COASTLINE
        if resolution is None:
            pass
        elif resolution in GSHSS_coast_resolutions:
            try:
                coast =  GSHHSFeature(scale=resolution)
            except Exception:
                print(f"Couldn't find/download GSHHSFeature {resolution} coastline")
        elif resolution in NE_coast_resolutions:
            try:
                coast = NaturalEarthFeature('physical', 'Coastline', scale=resolution)
            except Exception:
                print(f"Couldn't find/download NaturalEarthFeature {resolution} coastline")
        else:
            print(f'Invalid coastline resolution: "{resolution}"')
        self.ax.add_feature(coast)


    def plot(self, lon, lat, sym='o', mec='k', ms=8, **kwargs):
        """
        plot a point

        :param lon:  longitude
        :param lat:  latitude
        :param sym:  symbol
        :param color: marker face color
        :param kwargs: keyword arguments to pass on to pyplot.plot()
        """
        self.ax.plot(lon, lat, sym,
                     transform=ccrs.PlateCarree(),
                     **kwargs)

    def arrow(self, x, y, dx, dy, **kwargs):
        """
        plot a arrow

        :param x:  longitude
        :param y:  latitude
        :param dx:  arrow length
        :param dy:  arrow length
        :param kwargs: keyword arguments to pass on to pyplot.arrow()
        """
        self.ax.arrow(x, y, dx, dy, transform=ccrs.PlateCarree(), **kwargs)

    def text(self, lon, lat, name, text_box=True,
                   ha='left', va='center', color='black', **kwargs):
        """
        plot a station label

        :param m: map object
        :param lon: longitude
        :param lat: latitude
        :param name: station name
        :param text_box: Put a semi-opaque box under the text
        :param ha: horizontal text alignment ('left', 'center', 'right')
        :param va: vertical text alignment ('center', 'middle', 'bottom')
        :param color: text color
        :param kwargs: keyword arguments to pass on to pyplot.text()
        """
        if text_box:
            kwargs['bbox'] = dict(boxstyle="round", pad=0.1, ec='none', fc='white', alpha=0.5)
        self.ax.text(lon, lat, name, va=va, ha=ha,
                     color=color, transform=ccrs.PlateCarree(), **kwargs)

    def save_map(self, filename, title, fontsize=12):
        """
        Saves map to a file

        :param filename:
        :param title: plot title
        """
        self.ax.set_title(title, fontsize=fontsize)
        self.fig.savefig(filename)


def _set_shade(a, intensity=None, cmap=cm.jet, shaded=True, **kwargs):
    '''
    sets shading for data array based on intensity layer or data value

    inputs:
        a - a 2-d array or masked array
        intensity - a 2-d array of same size as a, representing the intensity
                    layer. if none is given the data itself is used after
                    getting the hillshade values see hillshade for more details.
        cmap - a colormap (e.g matplotlib.colors.LinearSegmentedColormap
              instance)
        shaded: boolean calculate shading
        scale,azdeg,altdeg - parameters for hilshade function see there for
              more details
    output:
        rgb - an rgb set of the Pegtop soft light composition of the data and
                intensity can be used as input for imshow()
    based on ImageMagick's Pegtop_light:
    http://www.imagemagick.org/Usage/compose/#pegtoplight
    '''
    if intensity is None:
        if shaded is True:
            # hilshade the data
            intensity = _hillshade(a, **kwargs)
        else:
            intensity = np.ones(a.shape)
    else:
        # or normalize the intensity
        intensity -= np.nanmin(intensity)
        intensity /= np.nanmax(intensity)
    # get rgb of normalized data based on cmap
    rgb = cmap((a - np.nanmin(a)) / float(np.nanmax(a) - np.nanmin(a)))[:, :, :3]
    # form an rgb eqvivalent of intensity
    d = intensity.repeat(3).reshape(rgb.shape)
    # simulate illumination based on pegtop algorithm.
    rgb = 2 * d * rgb + (rgb**2) * (1 - 2*d)
    # make sure we are below 1
    rgb *= 0.999
    return rgb


def _hillshade(data, scale=1.0, azdeg=165.0, altdeg=45.0):
    '''
    Convert data to hillshade based on matplotlib.colors.LightSource class.

    input:
         data - a 2-d array of data
         scale - scaling value of the data. higher number = lower gradient
         azdeg - where the light comes from: 0 south ; 90 east ; 180 north ;
                      270 west
         altdeg - where the light comes from: 0 horizon ; 90 zenith
    output: a 2-d shading array (normalized 0 to 1)
    '''
    # gradient in x and y directions
    dx, dy = np.gradient(data / float(scale))
    # slope = np.pi/2. - arctan(hypot(dx, dy))
    slope = np.pi/2. - np.arctan(np.sqrt(dx*dx + dy*dy))
    aspect = np.arctan2(dx, dy)

    # convert alt, az to radians
    azimuth = np.radians(azdeg)
    altitude = np.radians(altdeg)
    intensity = np.sin(altitude) * np.sin(slope)\
        + np.cos(altitude) * np.cos(slope)\
        * np.cos((-azimuth - np.pi/2.) - aspect)
    # Scale from 0 to 1
    intensity -= np.nanmin(intensity)
    intensity /= np.nanmax(intensity)
    return intensity



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("seaplan/version.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="seaplan",
    version=version['__version__'],
    author="Wayne Crawford",
    author_email="crawford@ipgp.fr",
    description="Sea-going mission planning",
    long_description=long_description,
    long_description_content_type="text/markdown; charset=UTF-8",
    url="https://github.com/WayneCrawford/seaplan",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
          'scipy>=1.6.3',
          'pyproj>=3.1.0',
          'shapely>=1.7',
          'pyshp>=2.1',
          'matplotlib>=3.4',
          'fiona>=1.9.6',
          'cartopy>=0.22',
          'pyyaml>=3.0',
          'xarray>=v2022.03.0',
          'jsonschema>=3.2',
          'docutils>=0.20',
          'jsonref>=1.0'
      ],
    entry_points={
         'console_scripts': [
             'seaplan=seaplan.sea_plan:main',
             'seaplan-validate=seaplan.validate_json:_console_script'
         ]
    },
    python_requires='>=3.9',
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics"
    ),
    keywords='oceanography, marine, OBS'
)



#!/usr/bin/env python3
"""
Plotting routines for seaplan
"""
import math
from datetime import datetime

import numpy as np
from scipy.io import netcdf
from matplotlib import pyplot as plt
# from mpl_toolkits.basemap import Basemap
import cartopy.crs as ccrs
from pylab import cm

from .bathy_map import BathyMap

# CONSTANTS
font_size = 7
text_offset = 0.02  # degrees from station
text_bbox = dict(boxstyle="round", pad=0.1, edgecolor='none',
                 facecolor='white', alpha=0.5)


def plot_map(events, orig_stations, params):
    """
    Plot basemap

    :param orig_stations: stations in CSV file(?)
    :param params: seaplan parameter dictionary
    """
    mp = params['map']
    m = BathyMap(mp['bounds'], mp.get('bathy_map', False), mp['grid']['x'],
                 mp['grid']['y'])
    if 'bathy_map' in mp:
        if isinstance(mp['bathy_map'], str):
            m.plot_image(pastel=True, shaded=True)
            m.plot_contours()
        elif mp['bathy_map'] is True:
            m.ax.stock_img()
            
    m.plot_coastlines(mp['coast_resolution'])
    
    for name, station in orig_stations.items():
        ms = 5
        if station['type'] == 'WayPoint':
            sym = 'k.'
        elif station['type'] == 'Operation':
            continue
        elif station['type'] == 'Survey':
            continue
        elif station['type'] == 'BB':
            sym = 'r+'
            ms = 10
        else:
            sym = 'r+'
        m.plot(station['lon'], station['lat'], sym, markersize=ms)
        # m.plot(event['lon'],event['lat'],'k+',latlon=True)
        # x, y = m(station['lon'] + text_offset, station['lat'])
        plt.text(station['lon']+ text_offset, station['lat'],
                 name, fontsize=font_size, va='center', color='gray')

    iLeg = 0
    # COLOR FOR EVERY "NEWLEG"
    legSyms = ['k-', 'r-', 'b-', 'g-', 'm-', 'c-', 'k--', 'r--', 'b--']
    prev_event = events[0].copy()
    for event in events:
        if event['station'] in orig_stations:
            cross_visited_stations(m, event)
        # Update leg # on "NEWLEG" event action
        if 'action' in event:
            if event['action'] == "NEWLEG":
                iLeg += 1
        if (params['map']['plot_past_tracks'] or
                (event['arrival']['time'] > datetime.now())):
            plot_track(m, event, prev_event, legSyms[iLeg % len(legSyms)])
        plot_event(m, event)
        prev_event = event.copy()
    dt = events[-1]['arrival']['time'] - events[0]['departure']['time']
    fig, base_name = close_map(params, dt)
    print(f'{base_name=}')
    fig.savefig(f'{base_name}.png', dpi=150.)



def plot_track(m, event, prev_event, sym):
    """
    Plot ship track
    """
    if event['dist_from_previous'] > 0:
        m.plot([event['lon'], prev_event['lon']],
                 [event['lat'], prev_event['lat']], sym)
        m.arrow(prev_event['lon'], prev_event['lat'],
                  .6 * (event['lon'] - prev_event['lon']),
                  .6 * (event['lat'] - prev_event['lat']),
                  head_width=0.015, length_includes_head=True)


def cross_visited_stations(m, event):
    """
    plot a cross if departure_time is specified & before current time
    """
    if 'depart_time' in event:
        if (datetime.strptime(event['depart_time'], '%Y-%m-%dT%H:%M')
                < datetime.now()):
            # x, y = m(event['lon'], event['lat'])
            m.plot(event['lon'], event['lat'], 'kx', markersize=12, linewidth=3)


def plot_event(m, event, debug=False):
    """
    Plot one mission event

    :param m: global coordinate system
    :param event: event dictionary
    """
    if debug:
        print("In plot_event()")
    if event['type'] == 'Operation':
        m.plot(event['lon'], event['lat'], 'r+', markersize=8, lw=5.0)
        # x, y = m(event['lon'], event['lat'] - text_offset)
        m.text(event['lon'], event['lat'], event['station'],
               fontsize=font_size, va='top', ha='center', color='red',
               bbox=text_bbox)
    else:
        # Station or WayPoint
        if event['type'] == 'WayPoint':
            m.plot(event['lon'], event['lat'], '+', markersize=5, lw=3.0)
        else:
            m.plot(event['lon'], event['lat'], 'ro', markersize=5, lw=3.0)
        if event['hours'] != 0:
            m.text(event['lon'] + text_offset, event['lat'],
                   event['station'], fontsize=font_size, va='center',
                    color='black', bbox=text_bbox)


def close_map(params, timedelta_total):
    """
    Saves map to a file

    :param params: seaplan parameters
    :param timedelta_total: total time spent (datetime.timedelta)
    :returns: figure object, base_name of output file
    """
    base_name = params['file'].split('.')[0]
    plt.title('{}: {:.0f} days {:.0f}h (speed={:g}, latency={:g}h)'.format(
              base_name, timedelta_total.days,
              timedelta_total.seconds / 3600.,
              params['timing']['ship_speed.kn_i'],
              params['timing']['ship_latency.h']),
              fontsize=12)
    fig1 = plt.gcf()    # Needed to save after "show" (which creates new fig)
    if params['map']['show_plot']:
        plt.show()
    return fig1, base_name



#!/usr/bin/env python3
"""Shading for python maps.  Included here for distribution"""

from pylab import gradient, sin, cos, arctan2, arctan, cm, pi, hypot


def set_shade(a, intensity=None, cmap=cm.jet, scale=10.0, azdeg=165.0,
              altdeg=45.0):
    '''
    sets shading for data array based on intensity layer or data value

    inputs:
        a - a 2-d array or masked array
        intensity - a 2-d array of same size as a (no check on that)
                    representing the intensity layer. if none is given
                    the data itself is used after getting the hillshade values
                    see hillshade for more details.
        cmap - a colormap (e.g matplotlib.colors.LinearSegmentedColormap
              instance)
        scale,azdeg,altdeg - parameters for hilshade function see there for
              more details
    output:
        rgb - an rgb set of the Pegtop soft light composition of the data and
                intensity can be used as input for imshow()
    based on ImageMagick's Pegtop_light:
    http://www.imagemagick.org/Usage/compose/#pegtoplight
    '''
    if intensity is None:
        # hilshading the data
        intensity = hillshade(a, scale=scale, azdeg=azdeg, altdeg=altdeg)
    else:
        # or normalize the intensity
        intensity = (intensity - intensity.min())\
                    / (intensity.max() - intensity.min())
    # get rgb of normalized data based on cmap
    rgb = cmap((a - a.min()) / float(a.max() - a.min()))[:, :, :3]
    # form an rgb eqvivalent of intensity
    d = intensity.repeat(3).reshape(rgb.shape)
    # simulate illumination based on pegtop algorithm.
    rgb = 2 * d * rgb + (rgb**2) * (1 - 2*d)
    return rgb


def hillshade(data, scale=10.0, azdeg=165.0, altdeg=45.0):
    '''
    Convert data to hillshade based on matplotlib.colors.LightSource class.

    input:
         data - a 2-d array of data
         scale - scaling value of the data. higher number = lower gradient
         azdeg - where the light comes from: 0 south ; 90 east ; 180 north ;
                      270 west
         altdeg - where the light comes from: 0 horizon ; 90 zenith
    output: a 2-d array of normalized hilshade
    '''
    # convert alt, az to radians
    az = azdeg * pi / 180.0
    alt = altdeg * pi / 180.0
    # gradient in x and y directions
    dx, dy = gradient(data / float(scale))
    slope = 0.5 * pi - arctan(hypot(dx, dy))
    aspect = arctan2(dx, dy)
    intensity = sin(alt) * sin(slope)\
        + cos(alt) * cos(slope) * cos(-az - aspect - 0.5 * pi)
    intensity = (intensity - intensity.min())\
        / (intensity.max() - intensity.min())
    return intensity



"""
Mission planning based on station and parameter filess

Reads in a YAML-format parameter file containing parameters and a list of events.
Details on this file are given in the README.rst file

The parameter file can also specify a "station file" containing the names,
latitudes and longitudes of the stations.  All of these stations will be
plotted on the map and if one of the events has the same name and no specified
latitude and longitude.

Outputs a table of events with predicted (or specified) times, in HTML and
RestructuredText formats.

Miscellaneous additional features:
  * Crosses out visited stations (departure_time before datetime.now()

  * Changes the color of the ship's track at every "NEWLEG" action

  * Plots a color bathymetric image if you specify map:bathy_map as:
    * a string: a NetCDF file at the given path
    * True: A low-res global image
"""
import csv
import sys
import argparse
from pathlib import Path
from datetime import datetime, timedelta
import math as m

import yaml
import docutils.core

from .sea_plot import plot_map
from .sea_print import event_table, print_used_stations
from .validate_json import validate


def main(debug=False):
    """
    Create an at-sea deployment plan
    """
    # Read from command line
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('parameter_file', metavar='PARAM_FILE',
                        help='YAML file containing all parameters')
    parser.add_argument('--no_map', action='store_true',
                        help="Don't print/save map")
    args = parser.parse_args()

    # Read in parameters and station positions
    validate(args.parameter_file)
    params, events = read_param_file(args.parameter_file)
    # print(f'{params=}')
    # print(f'{events=}')
    orig_stations = read_station_file(params['station_file'])
    # print(f'{orig_stations=}')

    # Set up variables
    total_hours = 0
    iLeg=0
    used_stations = []
    if debug:
        print(orig_stations)
    stations = orig_stations.copy()

    # Setup "past" information
    events[0]['dist_from_previous'] = 0
    prev_event = events[0].copy()
    prev_event, _ = _handle_existence(prev_event, stations, params)

    if 'depart_time' in prev_event:
        depart_base = datetime.strptime(prev_event['depart_time'],
                                        '%Y-%m-%dT%H:%M')
    else:
        depart_base = datetime.strptime(prev_event['arrive_time'],
                                        '%Y-%m-%dT%H:%M')

    # Loop over events
    for event in events:
        # Handle pre-existence of stations
        event, stations = _handle_existence(event, stations, params)
        if event_station(event, orig_stations):
            used_stations.append(event['station'])
        # Calculate distance from previous event
        event['dist_from_previous'] = distance(event, prev_event)
        # Update leg # on "NEWLEG" event action
        if 'action' in event:
            if event['action'] == "NEWLEG":
                iLeg += 1
        # Calculate time on station, latency, and total times
        event, total_hours = calc_times(event, params['timing'], depart_base,
                                        total_hours)
        prev_event = event.copy()
    # Print and save event table
    t = event_table(events, args.parameter_file, params, 'RST')
    print(t)
    outfname = Path(args.parameter_file).with_suffix('.rst').name
    with open(outfname, 'w') as f:
        f.write(t)
    htmlfname = Path(args.parameter_file).with_suffix('.html').name

    docutils.core.publish_file(
        source_path=outfname,
        destination_path=htmlfname,
        writer_name="html")

    # Print list of CSV-file stations used
    print_used_stations(orig_stations, used_stations)

    # Make a map
    if not args.no_map:
        plot_map(events, orig_stations, params)


def _handle_existence(event, stations, params):
    """
    Handle pre-existence of a station for an event
    """

    # If the event is at an existing station, modify/complete the information
    event = event_from_station(event, stations, params)

    # print(f'{event=}')
    # print(f'{stations=}')
    if not event_station(event, stations):
        # If the event is not at an existing station, save its lat and lon
        try:
            stations[event['station']] = dict(lat=event['lat'],
                                              lon=event['lon'])
        except:
            print(f'Error, event "{event["station"]}" has no coordinates')
            sys.exit(2)
    return event, stations


def read_param_file(parameter_file):
    """
    READ IN PARAMETERS AND SEQUENTIAL LIST OF EVENTS FROM YAML FILE

    Variables that can be provided in the parameter files are
    (default in [braces]):
    parameters (also can be called 'variables' in v0.x)
        description:
            title (str): will be printed at top of HTML page
            notes (list of str): will be printed just below
        stations:
            file(str): CSV (semicolon-separated) file of
                       station;Lat;Lon;Type;Comment
            times: {type1: hours, type2:hours, ...}  # Hours to deploy one station
                                                     # of given type
        map:
            show_plot: Show the map of positions and tracks [False]
            bounds: [left, right, bottom, top]  Decimal degrees    [MANDATORY if
                    plot_map==True]
            bathy (str): netCDF file to plot as background image on the map
                  (True): Plot an etopo image as background
                  (False): No background map
            plot_past_tracks: Plot past ship tracks [True]
        printout:
            show_comments : Show comments in output [False]
            print_past : Print out past events [False]
            print_waypoints: Print out waypoints (events with 0 time) [True]
        timing:
            ignore_arrive_depart: Ignore arrive_time and depart_time to test time
                                  predictions [False]
            ship_speed.kn_i: Cruising speed of the ship in knots [10.]
            ship_latency: Hours lost getting to speed from stopped [0.5]
    events: Sequential list of events to perform [MANDATORY].
                Each item must have a "station" key (str) and may have the
                following keys:
                    action(str): the action to perform ("deploy",
                                 "recover","survey","waypt"...)
                    lat (float): site latitude  [station.lat or 0]
                    lon (float): site longitude  [station.lat or 0]
                    arrive_time (ISO):  [previous depart_time + transit
                                         + latency]
                    depart_time (ISO):  [arrive_time + hours]
                    hours (float): hours expected for operation  [0 or
                                   stations[type].times]
                    comment (str): comments
                    speed (float): speed between previous and given site
                                  [ship_speed.kn_i]
    """
    stream = open(parameter_file, 'r')
    p = yaml.safe_load(stream)
    stream.close()
    event_list = p['events']

    # SET DEFAULT VALUES
    params = {'file': parameter_file,
              'description': {'title': 'SEAPLAN',
                              'notes': []},
              'station_file': {'name': False, 'field_separator': ';'},
              'printout': {'show_comments': False,
                           'print_past': False,
                           'print_waypoints': True},
              'timing': {'ignore_depart_arrive_times': False,
                         'ship_speed.kn_i': 10,
                         'ship_latency.h': 0.5,
                         'default_action_times.h': {}},
              'map': {'show_plot': False,
                      'bounds': [],
                      'grid': {'x': 1, 'y': 1},
                      'bathy_map': False,
                      'coast_resolution': 'auto',
                      'plot_past_tracks': True}}
    # UPDATE WITH ENTERED VALUES
    for key in params.keys():
        if 'parameters' in p and 'variables' in p:
            ValueError("Can't have both 'parameters and 'variables' in the"
                       "YAML file (prefer 'parameters')")
        elif 'parameters' in p:
            p_input = p['parameters']
        elif 'variables' in p:
            warnings.warn('"variables" is grandfathered, use "parameters"')
            p_input = ['variables']
        else:
            ValueError("Missing top-level 'parameters' in the YAML file")
        if key in p_input:
            params[key].update(p_input[key])

    return params, event_list


def _verify_fields(field_order):
    if 'name' not in field_order:
        print('Error: no "name" in 1st line of csv file')
        print(field_order)
        sys.exit(1)
    if 'lon' not in field_order:
        if not ('lond' in field_order and 'lonm' in field_order):
            print('Error: no "lon" or "lond"&"lonm" in 1st line of csv file')
            sys.exit(1)
    if 'lat' not in field_order:
        if not ('latd' in field_order and 'latm' in field_order):
            print('Error: no "lat" or "latd"&"latm" in 1st line of csv file')
            sys.exit(1)


def _deg_min_to_deg(degrees, minutes):
    return m.sign(float(degrees)) * (m.abs(float(degrees))
                                     + m.abs(float(minutes)/60.))


def read_station_file(station_file, debug=False):
    """  READ STATION POSITONS FROM CSV FILE  """
    stations = dict()
    if station_file['name']:
        with open(station_file['name']) as csvfile:
            reader = csv.reader(csvfile,
                                delimiter=station_file['field_separator'])
            field_order = next(reader)
            field_order = [x.lower() for x in field_order]
            _verify_fields(field_order)

            for row in reader:
                i = 0
                # Read into appropriate fields
                temp = dict()
                for field in field_order:
                    temp[field] = row[i]
                    i += 1
                if debug:
                    print(temp.keys())
                if 'lat' not in temp:
                    temp['lat'] = _deg_min_to_deg(temp['latd'], temp['latm'])
                if 'lon' not in temp:
                    temp['lon'] = _deg_min_to_deg(temp['lond'], temp['lonm'])

                # Fill station dictionary
                station = dict(lat=float(temp['lat']), lon=float(temp['lon']))
                station['type'] = temp.get('type', '')
                station['comment'] = temp.get('comment', '')

                # Put into stations dictionary
                stations[temp['name']] = station
    else:
        print(f'{station_file=} has no `name`, returning empty dict')
        stations = dict()
    return stations


def _degrees_to_kilometers(degrees, radius=6371):
    """
    Convenience function to convert (great circle) degrees to kilometers
    assuming a perfectly spherical Earth.  Taken from obspy

    :type degrees: float
    :param degrees: Distance in (great circle) degrees
    :type radius: int, optional
    :param radius: Radius of the Earth used for the calculation.
    :rtype: float
    :return: Distance in kilometers as a floating point number.

    .. rubric:: Example

    >>> _degrees_to_kilometers(1)
    111.19492664455873
    """
    return degrees * (2.0 * radius * m.pi / 360.0)


def _locations_to_degrees(lat1, long1, lat2, long2):
    """
    Convenience function to calculate the great circle distance between two
    points on a spherical Earth.

    This method uses the Vincenty formula in the special case of a spherical
    Earth. Taken from obspy.

    :type lat1: float
    :param lat1: Latitude of point 1 in degrees
    :type long1: float
    :param long1: Longitude of point 1 in degrees
    :type lat2: float
    :param lat2: Latitude of point 2 in degrees
    :type long2: float
    :param long2: Longitude of point 2 in degrees
    :rtype: float or :class:`numpy.ndarray`
    :return: Distance in degrees as a floating point number

    .. rubric:: Example

    >>> _locations_to_degrees(5, 5, 10, 10)
    7.0397014191753815
    """
    # Convert to radians.
    lat1 = m.radians(lat1)
    lat2 = m.radians(lat2)
    long1 = m.radians(long1)
    long2 = m.radians(long2)
    long_diff = long2 - long1
    gd = m.degrees(
        m.atan2(
            m.sqrt((
                m.cos(lat2) * m.sin(long_diff)) ** 2 +
                (m.cos(lat1) * m.sin(lat2) - m.sin(lat1) *
                    m.cos(lat2) * m.cos(long_diff)) ** 2),
            m.sin(lat1) * m.sin(lat2) + m.cos(lat1) * m.cos(lat2) *
            m.cos(long_diff)))
    return gd


def distance(site1, site2):
    """
    Distance in nautical miles between two sites

    Sites must have 'lat' and 'lon' keys and values in decimal degrees
    """
    # Why are these two cases?  In case nothing was entered?  Should be None!
    # if (site1['lat'] == 0 and site1['lon'] == 0)
    #         return 0
    # if (site2['lat'] == 0 and site2['lon'] == 0):
    #         return 0
    degrees = _locations_to_degrees(site1['lat'], site1['lon'],
                                    site2['lat'], site2['lon'])
    nm = _degrees_to_kilometers(degrees) / 1.852
    return nm


def _default_hours(event, params):
    if 'timing' in params:
        if 'default_action_times.h' in params['timing']:
            at_h = params['timing']['default_action_times.h']
            if 'action' in event:
                if event['action'] in at_h:
                    action = event['action']
                    if 'type' in event:
                        if event['type'] in at_h[action]:
                            return at_h[action][event['type']]
                        else:
                            return at_h[action].get('default', 0.)
                    else:
                        return at_h[action].get('default', 0.)
    return 0.


def event_from_station(event, stations, params):
    """
    Find the station corresponding to event['station']
    """
    station = event_station(event, stations)
    if station:
        if 'lat' in event:
            print(f'"{event["station"]}" latitude ignored, using previously '
                  'specified value')
        elif 'lon' in event:
            print(f'"{event["station"]}" longitude ignored, using previously '
                  'specified value')
        event['lat'] = station['lat']
        event['lon'] = station['lon']
        event['type'] = event.get('type', station.get('type', ''))
        event['hours'] = event.get('hours', _default_hours(event, params))
        event['comment'] = add_comments(event.get('comment', ''),
                                        station.get('comment', ''))
    else:
        event['hours'] = event.get('hours', 0.)
        if event['hours'] == 0:
            event['type'] = event.get('type', 'Waypoint')
        else:
            event['type'] = event.get('type', 'Operation')
        event['comment'] = add_comments(event.get('comment', ''), '')
    return event


def event_station(event, stations, debug=True):
    """ Find the station corresponding to event['station'] """
    if event['station'] in stations:
        return stations[event['station']]
    return None


def choose_time(input_time, forced_time, ignore_forced_times):
    """
    Return the time to associate with an arrival or departure

    input_time :    a calculated time
    forced_time : a hand-entered time (may be empty)
    ignore_forced times: whether to used the forced time or not
    """
    if forced_time and not ignore_forced_times:
        return dict(time=datetime.strptime(forced_time, '%Y-%m-%dT%H:%M'),
                    forced=True)
    return dict(time=input_time, forced=False)


def calc_times(event, tparams, depart_base, total_hours):
    """ Calculate arrival and departure times

    :type event: dict (should it become a class?)
    :param event: seaplan event
    :type tparams: dict
    :param tparams: timing parameters
    :type depart_base: datetime
    :param depart_base: base depart time
    :type total_hours: float
    :param total_hours: total hours for the event
    :rtype: dict
    :return: Event with recalculated times
    """
    time_on_station = event['hours']
    if time_on_station == 0:
        latency = 0
    else:
        latency = tparams['ship_latency.h']
    ship_speed = event.get('speed', tparams['ship_speed.kn_i'])
    imposed_arrive_time = event.get('arrive_time', False)
    imposed_depart_time = event.get('depart_time', False)
    event['transit_hours'] = event['dist_from_previous'] / ship_speed + latency
    arrival = depart_base + timedelta(seconds=3600 *
                                      (total_hours + event['transit_hours']))
    event['arrival'] = choose_time(arrival, imposed_arrive_time,
                                   tparams['ignore_depart_arrive_times'])
    departure = event['arrival']['time'] + timedelta(seconds=3600
                                                     * time_on_station)
    event['departure'] = choose_time(departure, imposed_depart_time,
                                     tparams['ignore_depart_arrive_times'])

    dt = event['departure']['time'] - depart_base
    total_hours = dt.days * 24 + dt.seconds/3600.

    return event, total_hours


def add_comments(event_comment, station_comment):
    """Combine event and station comments"""

    if station_comment and event_comment:
        # Put "e:" before the event_comment and "s:" before the station_comment
        return f'e:{event_comment}; s:{station_comment}'
    elif station_comment:
        return station_comment
    elif event_comment:
        return event_comment
    else:
        return ''


if __name__ == '__main__':
    sys.exit(main())
